##############################################################################
# == Configuration file for the Sphinx documentation builder =================
#####
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from __future__ import annotations

import os
import sys
from datetime import datetime
from pathlib import Path

from sphinx.application import Sphinx

DOC_ROOT = Path(__file__).parent.resolve()
PROJECT_ROOT = DOC_ROOT.parent
SRC_ROOT = PROJECT_ROOT.joinpath("src")
DOC_EXT_ROOT = DOC_ROOT.joinpath("_ext")

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

source_paths = [
    *os.getenv("SPHINX_APIDOC_SOURCES", "").split(":"),
    DOC_ROOT,
    SRC_ROOT,
    DOC_EXT_ROOT,
]

for source_path in source_paths:
    if not isinstance(source_path, Path):
        source_path = Path(source_path)
    sys.path.insert(0, str(source_path.resolve()))

#####
##############################################################################

##############################################################################
# -- Project information -----------------------------------------------------
#####
from sphinx_terraform import __version__

_today = datetime.today()

project = os.getenv("SPHINX_CONF_PROJECT", "Sphinx-Terraform")
author = os.getenv("SPHINX_CONF_AUTHOR", "Charles Bouchard-Légaré")
copyright = f"{_today.year}, {author}"

# The full version, including alpha/beta/rc tags
release = os.getenv("SPHINX_CONF_RELEASE", __version__)
version = ".".join(release.split(".")[:2])

#####
##############################################################################


##############################################################################
# -- General configuration ---------------------------------------------------

# If your project needs a minimal Sphinx version, state it here.
needs_sphinx = "3"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",  # Google-style docstrings
    "sphinx.ext.intersphinx",  # Link to other Sphinx docs
    "sphinx.ext.autodoc",  # Generate doc from python source
    "sphinx_autodoc_typehints",  # Read types from type annotations
    "sphinx_paramlinks",  # Link to function parameters
    "sphinx.ext.viewcode",  # Add links to highlighted source code
    "sphinx.ext.todo",  # To do notes within the documentation
    # "sphinx.ext.extlinks",  # Markup to shorten external links
    "sphinx.ext.ifconfig",  # Conditional documentation parts
    "sphinx.ext.graphviz",  # Add Graphviz graphs
    "sphinx_copybutton",  # «Copy» to clipboard javascript button
    "sphinxcontrib.bibtex",  # BibTeX style citations and bibliography
    "sphinxcontrib.programoutput",  # Insert the output of arbitrary commands into documents.
    "sphinx_tabs.tabs",  # Create tabbed content when building HTML.
    "sphinx_terraform",
    "sphinx_terraform.viewcode",
    "myst_parser",
]
for module_file in DOC_EXT_ROOT.glob("[!_]*.py"):
    extensions.append(module_file.stem)


suppress_warnings = ["extlinks.hardcoded"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = [".rst"]

# The master toctree document without its file extension.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# Directories in which to search for additional message catalogs (see language),
# relative to the source directory. The directories on this path are searched
# by the standard gettext module.
#
# Internal messages are fetched from a text domain of sphinx; so if you add the
# directory ./locale to this setting, the message catalogs (compiled from .po
# format using msgfmt) must be in ./locale/language/LC_MESSAGES/sphinx.mo. The
# text domain of individual documents depends on gettext_compact.
locale_dirs = ["locales"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, "()" will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "friendly"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# A string of reStructuredText that will be included at the beginning of every
# source file that is read. This is a possible place to add substitutions that
# should be available in every file (another being rst_epilog).
rst_prolog = f"""
.. |project| replace:: **{project}**
"""

# A string of reStructuredText that will be included at the end of every source
# file that is read. This is a possible place to add substitutions that should
# be available in every file (another being rst_prolog).
# rst_epilog = ""

#####
##############################################################################


##############################################################################
# -- Builders configuration --------------------------------------------------
#####

###
# -- Options for linkcheck: the builder that checks all links goes somewhere
#
# A list of regular expressions that match URIs that should not be checked
# when doing a linkcheck build.
linkcheck_ignore = []

#
###

###
# -- Options for HTML output
#
# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "icon_links": [
        {
            "name": "GitLab",
            "url": "https://gitlab.com/cblegare/sphinx-terraform",
            "icon": "fab fa-gitlab",
        },
    ],
    "show_prev_next": True,
    "use_edit_page_button": True,
}

# A dictionary of values to pass into the template engine’s context for all
# pages. Single values can also be put in this dictionary using the
# -A command-line option of sphinx-build.
if os.getenv("CI_MERGE_REQUEST_IID", None):
    review_app = {
        "project_id": os.getenv("CI_PROJECT_ID"),
        "project_path": os.getenv("CI_PROJECT_PATH"),
        "mr_id": os.getenv("CI_MERGE_REQUEST_IID"),
        "server_url": os.getenv("CI_SERVER_URL"),
    }
else:
    review_app = None
html_context = {
    # "gitlab_url": "https://gitlab.com", # or your self-hosted GitLab
    "gitlab_user": "cblegare",
    "gitlab_repo": "sphinx-terraform",
    "gitlab_version": "main",
    "doc_path": "docs",
    # Enable the Giltab review app widget,
    # see also
    #   _templates/layout.html
    #   https://docs.gitlab.com/ee/ci/review_apps/#configuring-visual-reviews
    "review_app": review_app,
}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = project

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = "pixmap/logo.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = "pixmap/favicon.ico"

# Custom sidebar templates, must be a dictionary that maps document names to
# template names.
#
# The keys can contain glob-style patterns, in which case all matching documents
# will get the specified sidebars. (A warning is emitted when a more than one
# glob-style pattern matches for any document.)
#
# The values can be either lists or single strings.
# If a value is a list, it specifies the complete list of sidebar templates to
# include. If all or some of the default sidebars are to be included, they must
# be put into this list as well.
#
# The default sidebars (for documents that don’t match any pattern) are defined
# by theme itself.
# Builtin themes are using these templates by default:
#  - localtoc.html
#  - relations.html
#  - sourcelink.html
#  - searchbox.html
#
# Builtin sidebar templates that can be rendered are:
#  - localtoc.html – a fine-grained table of contents of the current document
#  - globaltoc.html – a coarse-grained table of contents for the whole documentation set,
#    collapsed
#  - relations.html – two links to the previous and next documents
#  - sourcelink.html – a link to the source of the current document, if enabled
#    in html_show_sourcelink
#  - searchbox.html – the “quick search” box
html_sidebars = {
    "**": [
        "search-field",
        "sidebar-nav-bs.html",
        # "localtoc.html",
        "sourcelink.html",
    ]
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# A list of CSS files.
# The entry must be a filename string or a tuple containing the filename
# string and the attributes dictionary. The filename must be relative to
# the html_static_path, or a full URI with scheme like http://example.org/style.css.
# The attributes is used for attributes of <link> tag.
# It defaults to an empty list.
html_css_files = [
    "style.css",
]

# If not "", a "Last updated on:" timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = "%b %d, %Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ""

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   "da", "de", "en", "es", "fi", "fr", "hu", "it", "ja"
#   "nl", "no", "pt", "ro", "ru", "sv", "tr"
# html_search_language = "en"

# A dictionary with options for the search language support, empty by default.
# Now only "ja" uses this config value
# html_search_options = {"type": "default"}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
# html_search_scorer = "scorer.js"

#
###

###
# -- Options for LaTeX output
#
#   For more information, see https://www.sphinx-doc.org/en/master/latex.html
#

# A dictionary that contains LaTeX snippets overriding those Sphinx usually
# puts into the generated .tex files. Its "sphinxsetup" key is described
# separately.
latex_elements = {
    # The paper size ("letterpaper" or "a4paper").
    # "papersize": "letterpaper",
    # The font size ("10pt", "11pt" or "12pt").
    # "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    # "preamble": "",
}

# We use Unicode character, so xelatex to the rescue.
latex_engine = "xelatex"

# This value determines how to group the document tree into LaTeX source files.
# It must be a list of tuples
#
#   (startdocname, targetname, title, author, theme, toctree_only),
#
# where the items are:
#
# startdocname
#   String that specifies the document name of the LaTeX file’s master
#   document. All documents referenced by the startdoc document in TOC trees
#   will be included in the LaTeX file. (If you want to use the default master
#   document for your LaTeX build, provide your master_doc here.)
#
# targetname
#   File name of the LaTeX file in the output directory.
#
# title
#   LaTeX document title. Can be empty to use the title of the startdoc
#   document. This is inserted as LaTeX markup, so special characters like a
#   backslash or ampersand must be represented by the proper LaTeX commands if
#   they are to be inserted literally.
#
# author
#   Author for the LaTeX document. The same LaTeX markup caveat as for title
#   applies. Use \\and to separate multiple authors, as in: "John \\and Sarah"
#   (backslashes must be Python-escaped to reach LaTeX).
#
# theme
#   LaTeX theme. See latex_theme.
#
# toctree_only
#   Must be True or False. If true, the startdoc document itself is not
#   included in the output, only the documents referenced by it via TOC trees.
#   With this option, you can put extra stuff in the master document that shows
#   up in the HTML, but not the LaTeX output.
latex_documents = [
    (
        master_doc,
        "{!s}-{!s}.tex".format(project, release),
        project,
        author,
        "manual",
        False,
    ),
]

#
###

#####
##############################################################################


##############################################################################
# -- Extension configuration -------------------------------------------------
#####

###
# -- Options for todo extension
#
# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True
###

###
# -- Options for Copybutto,
#
# Using regexp prompt identifiers
copybutton_prompt_text = (
    r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
)
copybutton_prompt_is_regexp = True
# Honor line continuation characters when copying multline-snippets
copybutton_line_continuation_character = "\\"
# Honor HERE-document syntax when copying multiline-snippets
copybutton_here_doc_delimiter = "EOT"
###

###
# -- Options for sphinxcontrib-bibtex
bibtex_bibfiles = [
    str(bibtext_file.relative_to(DOC_ROOT))
    for bibtext_file in DOC_ROOT.rglob("*.bib")
]
###

###
# -- Options for intersphinx extension ---------------------------------------
#
# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
    "gitlab": ("https://python-gitlab.readthedocs.io/en/stable/", None),
    "setuptools": ("https://setuptools.readthedocs.io/en/latest/", None),
}
#
nitpicky = True
nitpick_ignore = [
    ("py:class", "re.Pattern"),
    ("py:class", "Path"),
    ("py:class", "typing_extensions.Protocol"),
    ("py:class", "domain label"),  # this one confuses me a lot
    ("py:class", "sphinx.ext.intersphinx.normalize_intersphinx_mapping"),
    ("py:class", "sphinx.environment.NoUri"),
    ("py:exc", "sphinx.environment.NoUri"),
    ("py:class", "sphinx.highlighting.PygmentsBridge"),
    ("py:class", "sphinx.util.typing.OptionSpec"),
    ("py:class", "OptionSpec"),
    ("py:class", "sphinx.roles.XRefRole"),
    ("py:class", "XRefRole"),
    ("py:class", "RoleFunction"),
    ("py:class", "sphinx.directives.ObjectDescription"),
    ("py:class", "ObjectDescription"),
    ("py:class", "sphinx.domains.Index"),
    ("py:class", "Index"),
    ("py:class", "sphinx.domains.ObjType"),
    ("py:class", "sphinx.domains.IndexEntry"),
    ("py:class", "ObjType"),
    ("py:meth", "sphinx.roles.XRefRole.process_link"),
    ("py:meth", "sphinx.roles.XRefRole.result_nodes"),
    ("py:class", "docutils.nodes.Element"),
    ("py:class", "docutils.nodes.document"),
    ("py:class", "docutils.nodes.Node"),
    ("py:class", "docutils.nodes.system_message"),
]
###

###
# -- Options for extlinks ----------------------------------------------------
#
extlinks = {
    "dudir": (
        "http://docutils.sourceforge.net/docs/ref/rst/directives.html#%s",
        "",
    ),
    "release": (
        "https://gitlab.com/cblegare/sphinx-terraform/-/releases/%s",
        "Sphinx-Terraform v",
    ),
    "mr": (
        "https://gitlab.com/cblegare/sphinx-terraform/-/merge_requests/%s",
        "!",
    ),
    "issue": (
        "https://gitlab.com/cblegare/sphinx-terraform/-/issues/%s",
        "#",
    ),
    "commiter": ("https://gitlab.com/%s", "@"),
}
###

###
# -- Options for sphinx_terraform --------------------------------------
#
terraform_sources = {"terraform": str(DOC_ROOT.joinpath("demo/terraform"))}
###

#####
##############################################################################

##############################################################################
# -- Local extension ---------------------------------------------------------
#####
#
# The conf.py file can act directly as a Sphinx extension by defining a
# setup function that takes a Sphinx object as an argument.
def setup(app: Sphinx):
    ...


#####
##############################################################################



"""
This is **not** an API.
"""
from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Union

from sphinx.application import Sphinx
from sphinx.builders import Builder
from sphinx.environment import BuildEnvironment
from sphinx.errors import SphinxError
from sphinx.highlighting import PygmentsBridge
from sphinx.util.logging import getLogger

from sphinx_terraform.i18n import t__
from sphinx_terraform.sphinxapi import ExtensionMetadata

if TYPE_CHECKING or sys.version_info < (3, 8, 0):
    from importlib_metadata import PackageNotFoundError, version
else:
    from importlib.metadata import PackageNotFoundError, version


try:
    # _package_name = __name__
    _package_name = "sphinx_terraform"
    __version__ = str(version(_package_name))  # type: ignore
except PackageNotFoundError:
    # package is not installed
    __version__ = "(please install the package)"


log = getLogger(__name__)


DOMAIN_NAME = "tf"


class SphinxTerraformError(SphinxError):
    category = "Sphinx-Terraform error"


def setup(app: Sphinx) -> ExtensionMetadata:
    app.require_sphinx("3")

    app.add_config_value("terraform_sources", app.srcdir, "env", [str, dict])
    app.add_config_value("terraform_comment_markup", "", "env", [str])
    from sphinx_terraform.markup import TerraformDomain

    app.add_domain(TerraformDomain)

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def get_config_terraform_comment_markup(env: BuildEnvironment) -> str:
    configured = env.config.terraform_comment_markup

    return str(configured)


def get_config_terraform_sources(
    env: BuildEnvironment,
) -> Dict[str, Path]:
    configured = env.config.terraform_sources

    if not configured or not isinstance(configured, (str, Dict, Path)):
        raise SphinxTerraformError(
            t__(
                "No terraform sources where configured in conf.py. "
                "Please provide a value to 'terraform_sources'."
            )
        )

    if isinstance(configured, (str, Path)):
        configured = {Path(configured).name: configured}

    for name in configured:
        path = Path(configured[name])
        if not path.is_absolute():
            path = Path(env.project.srcdir, path)
        configured[name] = path

    return configured


def get_env(app_or_env: Union[Sphinx, BuildEnvironment]) -> BuildEnvironment:
    if isinstance(app_or_env, BuildEnvironment):
        return app_or_env
    if isinstance(app_or_env.env, BuildEnvironment):
        return app_or_env.env
    raise SphinxTerraformError("Build environment not ready.")


def get_app(app_or_env: Union[Sphinx, BuildEnvironment]) -> Sphinx:
    if isinstance(app_or_env, BuildEnvironment):
        return app_or_env.app
    return app_or_env


def get_builder(app_or_env: Union[Sphinx, BuildEnvironment]) -> Builder:
    app = get_app(app_or_env)
    if isinstance(app.builder, Builder):
        return app.builder

    raise SphinxTerraformError("Builder not ready.")


def get_highlighter(
    app_or_env: Union[Sphinx, BuildEnvironment]
) -> PygmentsBridge:
    builder = get_builder(app_or_env)

    if hasattr(builder, "highlighter"):
        return builder.highlighter  # type: ignore

    raise SphinxTerraformError("Unsupported builder.")



"""
Register object types that exists in referenced documentation.

Intersphinx will only link to objects of types it knows, hence we
register some types, only for hyperlinking purposes.

See also:
    Taken from https://github.com/sphinx-doc/sphinx/blob/master/doc/conf.py
"""

from __future__ import annotations

import re

from sphinx import addnodes  # noqa

event_sig_re = re.compile(r"([a-zA-Z-]+)\s*\((.*)\)")


def setup(app):
    from sphinx.util.docfields import GroupedField

    app.add_object_type(
        "confval",
        "confval",
        objname="configuration value",
        indextemplate="pair: %s; configuration value",
    )

    fdesc = GroupedField(
        "parameter", names=("param",), label="Parameters", can_collapse=True
    )

    app.add_object_type(
        "event",
        "event",
        "pair: %s; event",
        parse_event,
        doc_field_types=[fdesc],
    )


def parse_event(env, sig, signode):
    m = event_sig_re.match(sig)
    if not m:
        signode += addnodes.desc_name(sig, sig)
        return sig
    name, args = m.groups()
    signode += addnodes.desc_name(name, name)
    plist = addnodes.desc_parameterlist()
    for arg in args.split(","):
        arg = arg.strip()
        plist += addnodes.desc_parameter(arg, arg)
    signode += plist
    return name



from __future__ import annotations

from typing import NamedTuple


class CodePosition(NamedTuple):
    """
    Define position in a code file, a pair of integers (line, column).
    """

    line: int
    column: int


class CodeSpan(NamedTuple):
    """
    Define a HCL definition within code, a signature and its position.
    """

    start_position: CodePosition
    end_position: CodePosition



"""
    sphinx.ext.extlinks
    ~~~~~~~~~~~~~~~~~~~

    Extension to save typing and prevent hard-coding of base URLs in the reST
    files.

    This adds a new config value called ``extlinks`` that is created like this::

       extlinks = {'exmpl': ('https://example.invalid/%s.html', caption), ...}

    Now you can use e.g. :exmpl:`foo` in your documents.  This will create a
    link to ``https://example.invalid/foo.html``.  The link caption depends on
    the *caption* value given:

    - If it is ``None``, the caption will be the full URL.
    - If it is a string, it must contain ``%s`` exactly once.  In this case the
      caption will be *caption* with the role content substituted for ``%s``.

    You can also give an explicit caption, e.g. :exmpl:`Foo <foo>`.

    Both, the url string and the caption string must escape ``%`` as ``%%``.

    :copyright: Copyright 2007-2022 by the Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from __future__ import annotations

import re
import warnings
from typing import Any, Dict, List, Tuple

import sphinx
from docutils import nodes, utils
from docutils.nodes import Node, system_message
from docutils.parsers.rst.states import Inliner
from sphinx.application import Sphinx
from sphinx.deprecation import RemovedInSphinx60Warning
from sphinx.locale import __
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util import logging
from sphinx.util.nodes import split_explicit_title
from sphinx.util.typing import RoleFunction

logger = logging.getLogger(__name__)


class ExternalLinksChecker(SphinxPostTransform):
    """
    For each external link, check if it can be replaced by an extlink.

    We treat each ``reference`` node without ``internal`` attribute as an external link.
    """

    default_priority = 500

    def run(self, **kwargs: Any) -> None:
        for refnode in self.document.findall(nodes.reference):
            self.check_uri(refnode)

    def check_uri(self, refnode: nodes.reference) -> None:
        """
        If the URI in ``refnode`` has a replacement in ``extlinks``,
        emit a warning with a replacement suggestion.
        """
        if "internal" in refnode or "refuri" not in refnode:
            return

        uri = refnode["refuri"]

        for alias, (base_uri, _caption) in self.app.config.extlinks.items():
            uri_pattern = re.compile(base_uri.replace("%s", "(?P<value>.+)"))
            match = uri_pattern.match(uri)
            if match and match.groupdict().get("value"):
                # build a replacement suggestion
                msg = __(
                    "hardcoded link %r could be replaced by an extlink "
                    "(try using %r instead)"
                )
                replacement = f":{alias}:`{match.groupdict().get('value')}`"
                logger.warning(
                    msg,
                    uri,
                    replacement,
                    location=refnode,
                    type="extlinks",
                    subtype="hardcoded",
                )


def make_link_role(name: str, base_url: str, caption: str) -> RoleFunction:
    # Check whether we have base_url and caption strings have an '%s' for
    # expansion.  If not, fall back the the old behaviour and use the string as
    # a prefix.
    # Remark: It is an implementation detail that we use Pythons %-formatting.
    # So far we only expose ``%s`` and require quoting of ``%`` using ``%%``.
    try:
        base_url % "dummy"
    except (TypeError, ValueError):
        warnings.warn(
            "extlinks: Sphinx-6.0 will require base URL to "
            "contain exactly one '%s' and all other '%' need "
            "to be escaped as '%%'.",
            RemovedInSphinx60Warning,
        )
        base_url = base_url.replace("%", "%%") + "%s"
    if caption is not None:
        try:
            caption % "dummy"
        except (TypeError, ValueError):
            warnings.warn(
                "extlinks: Sphinx-6.0 will require a caption string to "
                "contain exactly one '%s' and all other '%' need "
                "to be escaped as '%%'.",
                RemovedInSphinx60Warning,
            )
            caption = caption.replace("%", "%%") + "%s"

    def role(
        typ: str,
        rawtext: str,
        text: str,
        lineno: int,
        inliner: Inliner,
        options: Dict = {},
        content: List[str] = [],
    ) -> Tuple[List[Node], List[system_message]]:
        text = utils.unescape(text)
        has_explicit_title, title, part = split_explicit_title(text)
        full_url = base_url % part
        if not has_explicit_title:
            if caption is None:
                title = full_url
            else:
                title = caption % part
        pnode = nodes.reference(title, title, internal=False, refuri=full_url)
        return [pnode], []

    return role


def setup_link_roles(app: Sphinx) -> None:
    for name, (base_url, caption) in app.config.extlinks.items():
        app.add_role(name, make_link_role(name, base_url, caption))


def setup(app: Sphinx) -> Dict[str, Any]:
    app.add_config_value("extlinks", {}, "env")
    app.connect("builder-inited", setup_link_roles)
    app.add_post_transform(ExternalLinksChecker)
    return {"version": sphinx.__display_version__, "parallel_read_safe": True}



from __future__ import annotations

import sys

if sys.version_info < (3, 8, 0):
    from typing_extensions import runtime_checkable
else:
    from typing import runtime_checkable

from sphinx_terraform.terraform import HclSignature, TerraformResourceSignature


def test_protocol():
    a_signature = TerraformResourceSignature(
        "someprovider", "somekind", "somename"
    )

    checkable = runtime_checkable(HclSignature)

    assert isinstance(a_signature, checkable)



"""
Similar to :mod:`sphinx.ext.viewcode`.
"""
from __future__ import annotations

from pathlib import Path
from typing import Any, Iterator

from docutils import nodes
from docutils.nodes import Element, Node
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.builders import Builder
from sphinx.environment import BuildEnvironment
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util import status_iterator
from sphinx.util.logging import getLogger
from sphinx.util.nodes import make_refnode

from sphinx_terraform import __version__, get_builder
from sphinx_terraform.i18n import t_, t__
from sphinx_terraform.markup import TerraformDomain
from sphinx_terraform.sphinxapi import ExtensionMetadata, HtmlPage
from sphinx_terraform.terraform import make_identifier
from sphinx_terraform.viewcode.html import HtmlWriter

OUTPUT_DIRNAME = "_terraform_files"

log = getLogger(__name__)


def setup(app: Sphinx) -> ExtensionMetadata:
    app.require_sphinx("3")

    app.connect("env-purge-doc", env_purge_doc)
    app.connect("doctree-read", doctree_read)
    app.connect("html-collect-pages", collect_pages)
    app.add_post_transform(ViewcodeAnchorToggler)

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def env_purge_doc(app: Sphinx, env: BuildEnvironment, docname: str) -> None:
    domain = TerraformDomain.get_instance(app)
    domain.store.purge_usage(docname)


class viewcode_anchor(Element):
    """
    A sentinel node for viewcode anchors.

    Nodes of this type will be

    *   converted to anchors in supported builders or
    *   removed otherwise.

    This happens as a post transform phase.
    See also :class:`~ViewcodeAnchorToggler`.
    """


def doctree_read(app: Sphinx, doctree: Node) -> None:
    """
    Find HCL documentation and gather associated source code.

    Args:
        app:
            From the app we get the environment, our storage (cache).
        doctree:
            The document tree we will traverse in order to find HCL signatures.
    """
    domain = TerraformDomain.get_instance(app)

    writer = HtmlWriter(domain, Path(OUTPUT_DIRNAME))

    for signature_node in _gen_hcl_signatures(doctree):
        module = signature_node.get("module")
        signature = signature_node.get("signature", None)

        if module and signature:
            entry = domain.register(module, signature)
            signature_node += viewcode_anchor(
                reftarget=str(writer.make_pagename(entry.file)),
                refid=make_identifier(entry.signature),
                refdoc=domain.env.docname,
            )


class ViewcodeAnchorToggler(SphinxPostTransform):
    """
    Convert or remove :class:`~viewcode_anchor`.
    """

    default_priority = 100

    def run(self, **kwargs: Any) -> None:
        if is_supported_builder(self.app):
            self.convert_viewcode_anchors()
        else:
            self.remove_viewcode_anchors()

    def convert_viewcode_anchors(self) -> None:
        for node in self.document.findall(viewcode_anchor):
            anchor = nodes.inline("", t_("[source]"), classes=["viewcode-link"])
            refnode = make_refnode(
                get_builder(self.app),
                node["refdoc"],
                node["reftarget"],
                node["refid"],
                anchor,
            )
            node.replace_self(refnode)

    def remove_viewcode_anchors(self) -> None:
        for node in list(self.document.findall(viewcode_anchor)):
            node.parent.remove(node)


def is_supported_builder(app: Sphinx) -> bool:
    if not isinstance(app.builder, Builder):
        return False

    if app.builder.format != "html":
        return False
    elif app.builder.name == "singlehtml":
        return False
    elif (
        app.builder.name.startswith("epub")
        and not app.builder.config.viewcode_enable_epub
    ):
        return False
    else:
        return True


def collect_pages(app: Sphinx) -> Iterator[HtmlPage]:
    domain = TerraformDomain.get_instance(app)
    writer = HtmlWriter(domain, Path(OUTPUT_DIRNAME))

    tf_files = sorted(domain.store.get_documented_files())

    for tf_file in status_iterator(
        tf_files,
        t__("highlighting code..."),
        "blue",
        len(tf_files),
        app.verbosity,
        str,
    ):
        code_page = writer.gen_sourcecode_html(tf_file)
        yield code_page

    index_page = writer.gen_root_module_index()
    yield index_page


def _gen_hcl_signatures(node: Node) -> Iterator[addnodes.desc_signature]:
    def condition(candidate: Node) -> bool:
        return (
            isinstance(candidate, addnodes.desc)
            and candidate.get("domain") == "tf"
        )

    description_node: addnodes.desc
    for description_node in node.findall(condition=condition):
        signature_node: addnodes.desc_signature
        for signature_node in description_node.findall(  # type: ignore # error: <nothing> has no attribute "findall" ?!?
            addnodes.desc_signature
        ):
            yield signature_node



from __future__ import annotations

import re
import sys
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from sphinx.environment import BuildEnvironment
from sphinx.util.logging import getLogger

from sphinx_terraform import SphinxTerraformError, get_config_terraform_sources
from sphinx_terraform.code import CodePosition, CodeSpan
from sphinx_terraform.i18n import t__

if TYPE_CHECKING or sys.version_info < (3, 8, 0):
    from typing_extensions import Protocol
else:
    from typing import Protocol

log = getLogger(__name__)

_TerraformModule = TypeVar("_TerraformModule", bound="TerraformModule")


class TerraformModule(NamedTuple):
    """
    Represent a Terraform module.

    A Terraform module is a directory containing Terraform definition files.
    and other modules.

    This implementation extends the meaning of Terraform modules for our
    documentation purposes by giving the each module a :attr:`~name` and
    a :attr:`~fullname`.

    The top-most module is called the **root** module.

    The root module name comes from the documentation configuration. See
    :confval:`terraform_sources`.  Its *fullname* is the same as its
    name.

    Sub modules are named by their directory name and their *fullname* is
    derived from their parents names using slashes (``/``) as separators.

    Example:
        .. code-block:: text

            └── root_module/
                ├── main.tf
                ├── first_sub_module/
                │   ├── main.tf
                │   └── sub_sub_module/
                └── second_sub_module/
                    └── main.tf

        defines the following module fullnames:

        *   ``root_module``
        *   ``root_module/first_sub_module``
        *   ``root_module/first_sub_module/sub_sub_module``
        *   ``root_module/second_sub_module``

    .. tip:: In Terraform, **files** have no structural meaning.
    """

    root_name: str
    root_path: Path
    name: str
    path: Path
    fullname: str

    @classmethod
    def from_module_path(
        cls: Type[_TerraformModule],
        env: BuildEnvironment,
        module_path: Union[str, Path] = "",
    ) -> _TerraformModule:
        if not module_path:
            return cls.from_module_parts(env)
        else:
            root, *rest = Path(module_path).parts
            return cls.from_module_parts(env, root, "/".join(rest))

    @classmethod
    def from_module_parts(
        cls: Type[_TerraformModule],
        env: BuildEnvironment,
        root_module_name: Optional[str] = None,
        submodule: Union[str, Path] = "",
    ) -> _TerraformModule:
        root = cls._find_root(env, root_module_name)

        if not submodule:
            return root

        path = Path(submodule)
        if not path.is_absolute():
            path = root.path.joinpath(path)

        name = str(path.relative_to(root.path))
        fullname = f"{root.name}/{name}"

        return cls(root.name, root.path, name, path.resolve(), fullname)

    @classmethod
    def _find_root(
        cls: Type[_TerraformModule],
        env: BuildEnvironment,
        root_module_name: Optional[str] = None,
    ) -> _TerraformModule:
        sources = get_config_terraform_sources(env)
        if len(sources) == 1:
            root_name, root_path = next(iter(sources.items()))
        elif root_module_name is None:
            raise SphinxTerraformError(
                t__("Can't determine the proper terraform source to use.")
            )
        else:
            try:
                root_name = root_module_name
                root_path = sources[root_name]
            except KeyError as e:
                raise SphinxTerraformError(
                    t__(
                        "Unknown Terraform source '%s'. "
                        "Please review 'terraform_sources' in conf.py."
                    )
                    % root_module_name
                ) from e
        return cls(
            root_name,
            root_path.resolve(),
            root_name,
            root_path.resolve(),
            root_name,
        )

    def __str__(self) -> str:
        return self.fullname


class HclSignature(Protocol):
    """
    Signature of a Hashicorp Configuration Language (HCL) body element.

    This does not consider anything regarding a block's definition body,
    only its signature.  All signature **must** be defined on a single
    line of HCL code, as per the grammar.

    In the context of Terraform, which is a software that uses HCL by default
    for its configuration, HCL signatures are **unique in a given module**.
    Even sub module can have some signature equivalent to one of its parents
    ones without conflicts.

    Body elements can be one of the following.

    **Blocks**
        Blocks are multi line definitions.  Their pseudo-grammar is

        .. code-block:: bnf

            Block = Identifier (StringLit|Identifier)* "{" Newline Body "}" Newline;

        A block's *signature* is anything **before** the opening scope
        (``{``)

    **One line blocks**
        One line blocks are the same as blocks but on a single line,
        and thus can only have a simple definition body.  Their grammar is

        .. code-block:: bnf

            OneLineBlock = Identifier (StringLit|Identifier)* "{" (Identifier "=" Expression)? "}" Newline;

        A one line block's *signature* is anything **before** the opening
        scope (``{``), just like multi-line blocks.

    **Attributes**
        Attributes are simple key-value pairs.  Their grammar is

        .. code-block:: bnf

            Attribute = Identifier "=" Expression Newline;

        In this case, the signature would only include the left operand of
        the assignation, labeled ``Identifier`` above.

    In Terraform, the first **Identifier** of **blocks** and
    **one line blocks** is the signature's :attr:`~type` and zero (0) or
    more **labels** (:attr:`~labels`) based on the remaining **Identifiers**.
    **Attributes** have an implicit type and one (1) label.

    See also:
        `The HCL syntax`_

    .. _The HCL syntax: https://github.com/hashicorp/hcl/blob/main/hclsyntax/spec.md
    """

    @property
    def type(self) -> TerraformBlockType:
        """
        The Terraform object type (``resource``, ``data``, etc).
        """
        ...

    @property
    def labels(self) -> List[str]:
        """
        The object's labels, the amount of which depends on the :attr:`~type`.
        """
        ...

    def regex(self) -> Union[str, re.Pattern[str]]:
        """
        A regex that will reliably match the right code signature in a line.

        See also:
            The :func:`~regex` implementation that covers most cases.
        """
        ...

    def __repr__(self) -> str:
        """
        A string representation of
        Returns:

        """
        ...

    def __str__(self) -> str:
        """
        A string representation of
        Returns:

        """
        ...


def regex(self: HclSignature) -> Union[str, re.Pattern[str]]:
    """
    Return a regex that matches the signature reliably within a module.

    The regex will match a line of HCL code

    *   starting with 0 or more whitespaces,
    *   followed by an identifier without quotes
        (:attr:`sphinx_terraform.terraform.HclSignature.type`),
    *   followed by 1 or more non-newline whitespaces,
    *   followed by the signature's labels
        (:attr:`sphinx_terraform.terraform.HclSignature.labels`),
        possibily between double quotes `"` and
        separated by 1 or more non-newline whitespaces,
    *   followed by 1 or more non-newline whitespaces,
    *   followed by the `{` character and
    *   ending with the newline character.

    In Terraform, signatures are unique in a given module, hence this
    implementation is only reliable when scoped within a module.

    Args:
        self:
            The HCL block signature object we want to find in HCL code.

    Returns:
        Compiled regular expression.
    """
    # Use a double negation to match non-newline whitespace characters.
    # That is
    #   - Not
    #     - one of
    #       - Non-whitespace
    #       - Carriage return character
    #       - New line character
    non_newline_whitespace = r"[^\S\r\n]"

    signature_regex_inner = rf"{non_newline_whitespace}+".join(
        [
            # the type do not have quotes
            self.type.value,
            *[
                # labels can be between double quotes
                rf"(\"{label}\"|{label})"
                for label in self.labels
            ],
            # end with the block opening bracket
            "{.*",
        ]
    )

    # Enclose with accepted leading and trailing whitespace characters
    signature_regex = rf"^\s*{signature_regex_inner}$"
    return signature_regex


def _repr(self: HclSignature) -> str:
    """
    Make a valid Python code string that create an equivalent instance.
    """
    labels_within_quotes = [f"'{label}'" for label in self.labels]
    return (
        f"{self.__class__.__name__}("
        f"'{self.type}', "
        f"{', '.join(labels_within_quotes)}"
        f")"
    )


def _str(self: HclSignature) -> str:
    return f"{'.'.join(self.labels)}"


def make_identifier(
    signature: HclSignature, module: Optional[TerraformModule] = None
) -> str:
    """
    Create an URL friendly identifier string from a signature.

    This notion of identifier string does not exist in Terraform per se,
    since we do factor in the definituon's module.

    The below example illustrates how the identifier is built.

    Example:
        Given the following definitions within the module "mod":

        .. code-block:: terraform

            # in submodule "mod", in any file

            resource "null_resource" "some-name" {
            }

            data "null_data" "other-name" {
            }

        The identifiers would be

        .. code-block:: text

            mod/resource-null_resource.some-name

            mod/data-null_data.other-name

        >>> resource = TerraformResourceSignature("null", "resource", "some-name")
        >>> make_identifier(resource)
        'resource-null_resource.some-name'

    Args:
        signature:
            The HCL block signature for the definition we create an identifier.
        module:
            An optional module name.

    Returns:
        URL friendly identifier.
    """
    base_identifier = f"{signature.type.value}-{'.'.join(signature.labels)}"

    if module:
        return f"{module.name}/{base_identifier}"
    else:
        return base_identifier


class TerraformBlockType(Enum):
    DATA = "data"
    VARIABLE = "variable"
    MODULE = "module"
    OUTPUT = "output"
    PROVIDER = "provider"
    REQUIREMENT = "requirement"
    RESOURCE = "resource"


class HclDefinition(NamedTuple):
    """
    Define a HCL object.

    We use this because many definitions could have identical signature
    scattered across different modules (folders).
    """

    signature: HclSignature
    """
    Identify a definition, the signature is the definition *header*.
    """

    file: Path
    """
    HCL file where this definition is found.
    """

    doc_code: CodeSpan
    """
    HCL code section where this is documented.
    """

    signature_code: CodeSpan
    """
    HCL code section where this is defined.
    """

    body_code: CodeSpan
    """
    HCL code section where this is defined.
    """

    usages: Set[str]
    """
    Document name where this definition is referenced.
    """

    @property
    def code(self) -> CodeSpan:
        """
        Return where to find the whole code of this definition.

        We consider that the docstring, signature and body are contiguous.
        """
        return CodeSpan(
            self.doc_code.start_position, self.body_code.end_position
        )


class TerraformResourceSignature(NamedTuple):
    provider: str
    kind: str
    name: str

    @property
    def type(self) -> TerraformBlockType:
        return TerraformBlockType.RESOURCE

    @property
    def labels(self) -> List[str]:
        return [f"{self.provider}_{self.kind}", self.name]

    regex = regex  # type: ignore # Assigning methods is unsupported by mypy

    __repr__ = _repr  # type: ignore # Assigning methods is unsupported by mypy

    __str__ = _str  # type: ignore # Assigning methods is unsupported by mypy


class TerraformDataSignature(NamedTuple):
    provider: str
    kind: str
    name: str

    @property
    def type(self) -> TerraformBlockType:
        return TerraformBlockType.DATA

    @property
    def labels(self) -> List[str]:
        return [f"{self.provider}_{self.kind}", self.name]

    regex = regex  # type: ignore # Assigning methods in NamedTuple is unsupported by mypy

    __repr__ = _repr  # type: ignore # Assigning methods in NamedTuple is unsupported by mypy

    __str__ = _str  # type: ignore # Assigning methods is unsupported by mypy


class TerraformModuleSignature(NamedTuple):
    name: str

    @property
    def type(self) -> TerraformBlockType:
        return TerraformBlockType.MODULE

    @property
    def labels(self) -> List[str]:
        return [self.name]

    regex = regex  # type: ignore # Assigning methods is unsupported by mypy

    __repr__ = _repr  # type: ignore # Assigning methods is unsupported by mypy

    __str__ = _str  # type: ignore # Assigning methods is unsupported by mypy


class TerraformVariableSignature(NamedTuple):
    name: str

    @property
    def type(self) -> TerraformBlockType:
        return TerraformBlockType.VARIABLE

    @property
    def labels(self) -> List[str]:
        return [self.name]

    regex = regex  # type: ignore # Assigning methods is unsupported by mypy

    __repr__ = _repr  # type: ignore # Assigning methods is unsupported by mypy

    __str__ = _str  # type: ignore # Assigning methods is unsupported by mypy


class TerraformOutputSignature(NamedTuple):
    name: str

    @property
    def type(self) -> TerraformBlockType:
        return TerraformBlockType.OUTPUT

    @property
    def labels(self) -> List[str]:
        return [self.name]

    regex = regex  # type: ignore # Assigning methods is unsupported by mypy

    __repr__ = _repr  # type: ignore # Assigning methods is unsupported by mypy

    __str__ = _str  # type: ignore # Assigning methods is unsupported by mypy


_TerraformStore = TypeVar("_TerraformStore", bound="TerraformStore")


class TerraformStore:
    def __init__(self, data: Dict[Any, Any]) -> None:
        """
        Private constructor. Use :meth:`~from_build_env` instead.

        Args:
            data: Existing data, probably read from cache.
        """
        self.data: Dict[TerraformModule, ModuleData] = data

    @classmethod
    def initial_data(cls: Type[_TerraformStore]) -> Dict[Any, Any]:
        return defaultdict(ModuleData.new)

    def register(
        self,
        module: TerraformModule,
        signature: HclSignature,
        docname: str,
    ) -> HclDefinition:
        """
        Register a definition signature for a given module.

        This signals that a definition is documented. It makes sure we know
        about this HCL definition within the local cache, then registers
        the docname in its know documentation usages.

        Args:
            module:
                The module where this definition is found.  Multiple
                definitions should be unique within a module.
            signature:
                A signature identifies a definition.
            docname:
                The document where this signature is documented.

        Raises:
            sphinx_terraform.SphinxTerraformError: No definition could
                be found matching the signature within this module.

        Returns
            The registered definition.
        """
        hcl_definition = self.data[module].find_definition(module, signature)
        hcl_definition.usages.add(docname)
        return hcl_definition

    def purge_usage(self, usage_source: str) -> None:  # noqa
        for module in self.data.values():
            if not isinstance(module, TerraformModule):
                continue
            for signature in self.data[module].definitions.keys():
                entry = self.data[module].definitions[signature]
                if usage_source in entry.usages:
                    entry.usages.remove(usage_source)
                if not entry.usages:
                    del self.data[module].definitions[signature]

    def get_code(self, tf_file: Path) -> List[str]:
        module = self.get_module(tf_file)
        return self.data[module].get_code(tf_file)

    def get_module(self, tf_file: Path) -> TerraformModule:
        module_path = tf_file.parent
        for module in self.data:
            if module.path == module_path:
                return module
        else:
            raise SphinxTerraformError(f"No module found at '{module_path}'.")

    def get_definitions(
        self,
        module: Optional[TerraformModule] = None,
        tf_file: Optional[Path] = None,
    ) -> Dict[HclSignature, HclDefinition]:
        module = self.get_module(tf_file) if tf_file else module

        def gen_definitions(
            module: Optional[TerraformModule],
        ) -> Iterator[Tuple[HclSignature, HclDefinition]]:
            if module:
                yield from self.data[module].definitions.items()
            else:
                for module in self.data:
                    yield from self.data[module].definitions.items()

        def condition(entry: HclDefinition) -> bool:
            return entry.file == tf_file if tf_file else True

        return {
            signature: definition
            for signature, definition in gen_definitions(module)
            if condition(definition)
        }

    def get_module_files(self) -> List[Tuple[TerraformModule, Path]]:
        return sorted(
            [
                (module, filepath)
                for module in self.data
                for filepath in self.data[module].get_documented_files()
            ]
        )

    def get_documented_files(
        self, module: Optional[TerraformModule] = None
    ) -> Set[Path]:
        def gen_files() -> Iterator[Path]:
            if module:
                yield from self.data[module].get_documented_files()
            else:
                for module_data in self.data.values():
                    yield from module_data.get_documented_files()

        return set(tf_file for tf_file in gen_files())

    def get_documentation(self, definition: HclDefinition) -> List[str]:
        start_line = definition.doc_code.start_position.line
        end_line = definition.doc_code.end_position.line

        code_with_doc = self.get_code(definition.file)[start_line:end_line]

        documentation = extract_docstring_from_comment(code_with_doc)
        return documentation


_ModuleData = TypeVar("_ModuleData", bound="ModuleData")


class ModuleData(NamedTuple):
    """
    What we store in the build environment for a given Terraform module.

    Here is a JSON-like representation:

    .. code-block:: text

        {
            code: {
                Path: [
                    # lines of code
                ]
            },
            definitions: {
                HclBlockSignature: HclSignatureDefinition
            }
        }
    """

    code: Dict[Path, List[str]]
    """
    Cache of raw HCL code indexed by file path.
    """

    definitions: Dict[HclSignature, HclDefinition]
    """
    Found block definitions indexed by their signature.
    """

    @classmethod
    def new(cls: Type[_ModuleData]) -> _ModuleData:
        return cls(defaultdict(list), dict())

    def find_definition(
        self, module: TerraformModule, signature: HclSignature
    ) -> HclDefinition:
        """
        Look for a Terraform definition in all HCL files of a module.

        We use an internal cache that is pickled with the Sphinx environment.
        If the cache misses, we parse the code to find the definition.

        Args:
            signature:
                A signature identifies a definition.

        Raises:
            sphinx_terraform.SphinxTerraformError: No definition could
                be found matching the signature within this module.

        Returns:
            The found definition.
        """
        log.debug(f"Looking for definition of {repr(signature)}.")
        try:
            hcl_definition = self.definitions[signature]
            log.debug(f"Found definition of {repr(signature)} in cache.")
        except KeyError:
            for tf_file in module.path.glob("*.tf"):
                hcl_definition = self._find_definition(signature, tf_file)  # type: ignore
                if not hcl_definition:
                    continue
                log.debug(f"Caching definition of {repr(signature)}.")
                self.definitions[signature] = hcl_definition
                return hcl_definition
            else:
                raise SphinxTerraformError(
                    f"Definition not found for {repr(signature)} in module {module}."
                )
        return hcl_definition

    def get_definitions(
        self, tf_file: Optional[Path] = None
    ) -> Dict[HclSignature, HclDefinition]:
        """
        Make a mapping of known HCL definitions.

        Args:
            tf_file:
                Optionally filter for a specific file.

        Returns:
            The returned dictionary is created on each call, thus removing
            items from it won't remove them from this store.
        """

        def condition(entry: HclDefinition) -> bool:
            return entry.file == tf_file if tf_file else True

        return {
            signature: entry
            for signature, entry in self.definitions.items()
            if condition(entry)
        }

    def get_documentation(self, signature: HclSignature) -> List[str]:
        definition = self.definitions[signature]

        start_line = definition.doc_code.start_position.line
        end_line = definition.doc_code.end_position.line
        code_with_doc = self.get_code(definition.file)[start_line:end_line]

        documentation = extract_docstring_from_comment(code_with_doc)
        return documentation

    def get_code(self, tf_file: Path) -> List[str]:
        if tf_file not in self.code:
            log.debug(f"Putting code from {tf_file} in cache.")
            self.code[tf_file] = tf_file.read_text().splitlines()
        raw_code = self.code[tf_file]
        return raw_code

    def get_documented_files(self) -> Set[Path]:
        return set(entry.file for entry in self.definitions.values())

    def _find_definition(
        self, signature: HclSignature, tf_file: Path
    ) -> Optional[HclDefinition]:
        log.debug(f"Looking for definition of {repr(signature)} in files.")
        raw_code = self.get_code(tf_file)
        log.debug(f"Looking for definition of {repr(signature)} in {tf_file}.")
        found_code = self._find_definition_code(signature, raw_code)
        if not found_code:
            return None

        doc_code, signature_code, body_code = found_code
        log.debug(f"Found definition of {repr(signature)} in code.")
        hcl_definition = HclDefinition(
            signature,
            tf_file,
            doc_code,
            signature_code,
            body_code,
            set(),
        )
        return hcl_definition

    def _find_definition_code(
        self, signature: HclSignature, lines: List[str]
    ) -> Optional[Tuple[CodeSpan, CodeSpan, CodeSpan]]:
        signature_code = self._lookup_signature(signature, lines)

        if not signature_code:
            return None

        return self._find_whole_definition_code(signature_code, lines)

    def _find_whole_definition_code(
        self, signature_code: CodeSpan, lines: List[str]
    ) -> Tuple[CodeSpan, CodeSpan, CodeSpan]:
        # We found the signature, now let's try to climb up to the
        # last signature in order to find include a comment
        log.debug("Looking for beginning of definition, including comment.")
        begining_line = self._find_begining_of_documented_code(
            signature_code.start_position.line, lines
        )
        doc_code = CodeSpan(
            CodePosition(begining_line, 0),
            CodePosition(
                signature_code.start_position.line,
                signature_code.start_position.column,
            ),
        )
        log.debug(f"Found beginning of definition on line {begining_line}.")
        log.debug("Looking for definition body.")
        end_line = self._find_end_of_block(
            signature_code.start_position.line + 1, lines
        )
        body_code = CodeSpan(
            CodePosition(
                signature_code.end_position.line,
                signature_code.end_position.column,
            ),
            CodePosition(end_line, 0),
        )
        log.debug(f"Definition body ends on line {end_line}.")
        return doc_code, signature_code, body_code

    def _lookup_signature(
        self, signature: HclSignature, lines: List[str]
    ) -> Optional[CodeSpan]:
        signature_regex = signature.regex()
        for i, line_of_code in enumerate(lines):
            if re.match(signature_regex, line_of_code):
                log.debug(f"Found signature on line {i}.")
                signature_code = CodeSpan(
                    CodePosition(
                        i, len(line_of_code) - len(line_of_code.lstrip())
                    ),
                    CodePosition(i, len(line_of_code.rstrip())),
                )
                return signature_code
        else:
            # We could not find 'signature' in this code.
            return None

    def _find_begining_of_documented_code(
        self, start_line: int, lines: List[str]
    ) -> int:
        inspected_line = start_line - 1
        within_multiline_comment = False
        while inspected_line > 0:
            current_line = lines[inspected_line].strip()
            if current_line.startswith("*/"):
                log.debug(
                    f"Found the end of a multiline comment on line {inspected_line}."
                )
                within_multiline_comment = True
                inspected_line -= 1
                continue
            if within_multiline_comment and current_line.startswith("/*"):
                # we found the begining of a '/* ... */' comment.
                # That marks the start of the docstring, thus the end of
                # this search.
                log.debug(
                    f"Found the beginning of a multiline comment on line {inspected_line}."
                )
                return inspected_line
            if (
                current_line.startswith("#")
                or current_line.startswith("//")
                or within_multiline_comment
            ):
                # We found a single line comment.
                # Continue upward
                log.debug(
                    f"Found a single line comment on line {inspected_line}."
                )
                inspected_line -= 1
                continue
            # We found something else:
            # - empty line
            # - something that is not a comment
            log.debug(f"Found a non-comment line on line {inspected_line}.")
            return inspected_line + 1
        # We got to the top of the file. That's ok.
        log.debug("Went up to start of file.")
        return 0

    def _find_end_of_block(  # noqa
        self, start_line_within_block: int, all_lines: List[str]
    ) -> int:
        inspected_line = start_line_within_block
        # Within a block, there is one brace in the stack
        curly_brace_balance = 1

        while inspected_line < len(all_lines):
            log.debug(f"Looking for block delimiters on line {inspected_line}.")
            log.debug(f"Delimiter stack is {curly_brace_balance}")
            inspected_line_of_code = all_lines[inspected_line]
            for character in inspected_line_of_code:
                if character == "{":
                    curly_brace_balance += 1
                elif character == "}":
                    curly_brace_balance -= 1
                if curly_brace_balance == 0:
                    return inspected_line
            inspected_line += 1

        raise SphinxTerraformError(
            f"Found end of file when expecting closing scope opened before "
            f"line {start_line_within_block}"
        )


def extract_docstring_from_comment(  # noqa
    code_with_doc: List[str],
) -> List[str]:
    found_some_docs = False
    absolute_doc_indent = 0
    documentation = []
    for line in code_with_doc:
        if not found_some_docs:
            match = re.match(r"^[/\*\s#]* (?P<content>\S+.*)", line)
            if not match:
                continue
            content = match.group("content")
            absolute_doc_indent = len(line) - len(content)
            found_some_docs = True
        else:
            content = line[absolute_doc_indent:]
        if content and re.match(r"^[/\*\s#]+$", content):
            content = None

        if content is not None:
            documentation.append(content)
    return documentation



from __future__ import annotations

import re
import sys
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import sphinx.parsers
from docutils import nodes, utils
from docutils.nodes import Element, Node, system_message
from docutils.parsers.rst import directives
from docutils.statemachine import StringList
from sphinx import addnodes
from sphinx.addnodes import desc_signature
from sphinx.application import Sphinx
from sphinx.builders import Builder
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain, Index, IndexEntry, ObjType
from sphinx.environment import BuildEnvironment
from sphinx.roles import XRefRole
from sphinx.util.logging import getLogger
from sphinx.util.nodes import make_refnode
from sphinx.util.typing import OptionSpec

from sphinx_terraform import (
    SphinxTerraformError,
    get_config_terraform_comment_markup,
    get_env,
)
from sphinx_terraform.i18n import t_
from sphinx_terraform.sphinxapi import (
    SphinxDomainObjectDescription,
    SphinxGeneralIndexEntry,
)
from sphinx_terraform.terraform import (
    HclDefinition,
    HclSignature,
    TerraformBlockType,
    TerraformDataSignature,
    TerraformModule,
    TerraformModuleSignature,
    TerraformOutputSignature,
    TerraformResourceSignature,
    TerraformStore,
    TerraformVariableSignature,
    make_identifier,
)

if TYPE_CHECKING or sys.version_info < (3, 8, 0):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


log = getLogger(__name__)

T = TypeVar("T", bound=str)


def parser_for_markup(markup_language_id: str) -> Type[sphinx.parsers.Parser]:
    markup_language_id = markup_language_id.lower()
    if markup_language_id in ("md", "myst", "markdown"):
        try:
            from myst_parser.sphinx_parser import MystParser

            return MystParser
        except ImportError as e:
            raise SphinxTerraformError(
                "In order to use Markdown in Terraform comments, "
                "you need the MyST parser. "
                "Install Sphinx-Terraform with the 'markdown' extra by "
                "issuing the 'pip install sphinx-terraform[markdown]' command. "
                "Then, enable it in your extensions.  "
                "See https://myst-parser.readthedocs.io/ for details"
            ) from e
    return sphinx.parsers.RSTParser


class TerraformObjectDirective(ObjectDescription[str]):
    option_spec: OptionSpec = {
        "noindex": directives.flag,
        "rootmodule": directives.unchanged,
        "module": directives.unchanged,
        "markup": directives.unchanged,
    }

    allow_nesting = False

    @property
    def _domain(self) -> TerraformDomain:
        return TerraformDomain.get_instance(self.env)

    @property
    def module(self) -> TerraformModule:
        if hasattr(self, "_hcl_mod"):
            return getattr(self, "_hcl_mod")  # type: ignore

        root_module = self.options.get("rootmodule", None)
        module = self.options.get("module", "")

        return TerraformModule.from_module_parts(self.env, root_module, module)

    @property
    def hcl_sig(self) -> Optional[HclSignature]:
        if hasattr(self, "_hcl_sig"):
            return getattr(self, "_hcl_sig")  # type: ignore
        return None

    @property
    def hcl_def(self) -> Optional[HclDefinition]:
        if hasattr(self, "_hcl_def"):
            return getattr(self, "_hcl_def")  # type: ignore
        return None

    @property
    def terraform_block_type(self) -> TerraformBlockType:
        possible_prefix = f"{self.domain}:"
        if self.name.startswith(possible_prefix):
            block_type_name = self.name[len(possible_prefix) :]
        else:
            block_type_name = self.name

        return TerraformBlockType(block_type_name)

    def parse_signature(self, signature: str) -> HclSignature:
        log.debug(f"Parsing signature '{signature}'.")
        block_type = self.terraform_block_type
        log.debug(f"Looking for signature parser for '{block_type.value}'.")

        stripped = signature.strip()
        if "/" in stripped:
            module_path, stripped = stripped.rsplit("/", maxsplit=1)
            self._hcl_mod = TerraformModule.from_module_path(
                self.env, module_path
            )

        parser = self._get_directive_signature_parser(block_type)
        parsed = parser(stripped)
        return parsed

    def _get_directive_signature_parser(
        self, block_type: TerraformBlockType
    ) -> Callable[[str], HclSignature]:
        parser = cast(
            Callable[[str], HclSignature],
            getattr(self, f"_parse_{block_type.value}_signature"),
        )
        return parser

    def _parse_resource_signature(self, signature: str) -> HclSignature:
        provider_kind, name = signature.split(".", maxsplit=1)
        provider, kind = provider_kind.split("_", maxsplit=1)
        return TerraformResourceSignature(provider, kind, name)  # type: ignore # Method assignation to NamedTuple messes with Mypy

    def _parse_data_signature(self, signature: str) -> HclSignature:
        provider_kind, name = signature.split(".", maxsplit=1)
        provider, kind = provider_kind.split("_", maxsplit=1)
        return TerraformDataSignature(provider, kind, name)  # type: ignore # Method assignation to NamedTuple messes with Mypy

    def _parse_module_signature(self, signature: str) -> HclSignature:
        return TerraformModuleSignature(signature)  # type: ignore # Method assignation to NamedTuple messes with Mypy

    def _parse_output_signature(self, signature: str) -> HclSignature:
        return TerraformOutputSignature(signature)  # type: ignore  # Method assignation to NamedTuple messes with Mypy

    def _parse_variable_signature(self, signature: str) -> HclSignature:
        return TerraformVariableSignature(signature)  # type: ignore # Method assignation to NamedTuple messes with Mypy

    def make_signature_nodes(self, parsed_sig: HclSignature) -> List[Node]:
        """
        Make nice but semantic nodes for the HCL block signature.

        A HCL block signature is made of one type identifier followed by
        a number of label identifiers. The amount of label identifiers is
        specified and fixed for a given type identifier.  For instance,
        a Terraform ``resource`` has always 2 label identifiers, like so:

        .. code-block:: tf

            resource "google_storage_bucket" "some-bucket" {
                # body of the block definition
                # ...
            }

        Args:
            parsed_sig:
                At this point, the signature has already been parsed and
                we simply use it to generate the correct nodes.

        Returns:
            A list of nodes to append to the signature node.
        """
        signature_nodes: List[Node] = []

        # start with the Terraform definition block type (resource, data...)
        signature_nodes.append(
            addnodes.desc_annotation(
                rawsource=parsed_sig.type.value, text=parsed_sig.type.value
            )
        )

        if parsed_sig.labels:
            first_label, *other_labels = parsed_sig.labels
            signature_nodes.append(addnodes.desc_sig_space())
            signature_nodes.append(
                addnodes.desc_name(rawsource=first_label, text=first_label)
            )

            for label in other_labels:
                signature_nodes.append(addnodes.desc_sig_literal_char(".", "."))
                signature_nodes.append(
                    addnodes.desc_name(rawsource=label, text=label)
                )

        return signature_nodes

    def handle_signature(self, sig: str, signode: desc_signature) -> str:
        hcl_sig = self.parse_signature(sig)
        hcl_def = self._domain.register(self.module, hcl_sig)
        self._hcl_sig = hcl_sig
        self._hcl_def = hcl_def

        signode["module"] = self.module
        signode["signature"] = hcl_sig
        signode["definition"] = hcl_def

        signature_nodes = self.make_signature_nodes(hcl_sig)
        signode.extend(signature_nodes)

        return make_identifier(hcl_sig, self.module)

    def add_target_and_index(
        self, name: T, sig: str, signode: desc_signature
    ) -> None:
        """
        Add cross-reference IDs and entries to self.indexnode, if applicable.

        This method will add this signature name to the general index.

        Args:
            name:
                The identifier of this signature as returned by
                :meth:`~handle_signature`.
            sig:
                One signature string, which is the object name we are indexing.
            signode:
                The signature document node. It acts like the `term` part
                of a glossary entry.
        """
        # We include our unique identifier in the signature's ids, which
        # in turn can be used to render HTML and "id" attribute.
        signode["ids"].append(name)

        signature = self._hcl_sig

        generalindex_entry = SphinxGeneralIndexEntry(
            entrytype="single",
            entryname=str(signature),
            targetid=name,
            mainname=str(signature),
            key=None,
        )
        inode = addnodes.index(entries=[generalindex_entry])
        self.indexnode.append(inode)

    def transform_content(self, contentnode: addnodes.desc_content) -> None:
        hcl_def = self._hcl_def

        code = self._domain.store.get_documentation(hcl_def)

        for line in code:
            log.debug(line)

        comment_nodes = self._parse_terraform_comment(hcl_def)
        contentnode.extend(comment_nodes)

    def _local_code_url(self, hcl_def: HclDefinition) -> str:
        return f"{hcl_def.file}#L{hcl_def.doc_code.start_position.line}-L{hcl_def.doc_code.end_position.line}"

    def _parse_comment_with_external_parser(
        self, hcl_def: HclDefinition, markup_id: str
    ) -> List[Node]:
        parser_class = parser_for_markup(markup_id)
        parser = parser_class()
        parser.set_application(self.env.app)
        document = utils.new_document(
            self._local_code_url(hcl_def), self.state.document.settings
        )

        code = self._domain.store.get_documentation(hcl_def)

        parser.parse("\n".join(code), document)
        return document.children

    def _parse_comment_with_current_parser(
        self, hcl_def: HclDefinition
    ) -> List[Node]:
        document = utils.new_document(
            self._local_code_url(hcl_def), self.state.document.settings
        )

        code = self._domain.store.get_documentation(hcl_def)

        self.state.nested_parse(
            StringList(
                code,
                source=str(hcl_def.file),
            ),
            self.content_offset,
            document,
        )

        return document.children

    def _parse_comment_markdown_comment(
        self, hcl_def: HclDefinition
    ) -> List[Node]:
        return self._parse_comment_with_external_parser(hcl_def, "markdown")

    def _parse_comment_restructuredtext_comment(
        self, hcl_def: HclDefinition
    ) -> List[Node]:
        return self._parse_comment_with_external_parser(
            hcl_def, "restructuredtext"
        )

    def _parse_terraform_comment(self, hcl_def: HclDefinition) -> List[Node]:
        markup_id = self.options.get(
            "markup", get_config_terraform_comment_markup(self.env)
        )
        if markup_id:
            return self._parse_comment_with_external_parser(hcl_def, markup_id)
        else:
            return self._parse_comment_with_current_parser(hcl_def)


class TerraformCrossReferenceRole(XRefRole):
    """
    Define a Terraform object reference role.

    Cross referencing Terraform objects works alike crossreference to
    objects of the Python domain.

    The customization of a standard cross-reference can be done either
    by supplying constructor parameters or subclassing and overwriting
    the :meth:`sphinx.roles.XRefRole.process_link` and/or
    the :meth:`sphinx.roles.XRefRole.result_nodes` methods.
    """

    def process_link(
        self,
        env: BuildEnvironment,
        refnode: Element,
        has_explicit_title: bool,
        title: str,
        target: str,
    ) -> Tuple[str, str]:
        """
        Process link for a given cross-reference role.

        See also:
            The parent class method docstring is something like

                Called after parsing title and target text, and creating
                the reference node (given in *refnode*).  This method can
                alter the reference node and must return a new (or the same)
                ``(title, target)`` tuple.
        Args:
            env:
                Sphinx build environment.
            refnode:
                The created referenced node, which can be altered here.
            has_explicit_title:
                An explicit title in a role is when a display string is
                provided as part of the role's interpreted text. For example

                .. code-block: rst

                    :ref:`Here is an explicit title<some-reference-target>`

                would diplay an hyperlink to ``some-reference-target`` with
                ``Here is an explicit title`` as the link text.

                This value is also available as a instance member with the
                same name (``self.has_explicit_title``).
            title:
                The link title.
            target:
                The link target identifier.

        Returns:
            Title and target strings.
        """
        log.debug(f"Processing links for {self._role_string()}.")
        return super().process_link(
            env, refnode, has_explicit_title, title, target
        )

    def result_nodes(
        self,
        document: nodes.document,
        env: BuildEnvironment,
        node: Element,
        is_ref: bool,
    ) -> Tuple[List[Node], List[system_message]]:
        """
        Add general index nodes just before returning the finished xref nodes.

        See also:
            The parent class method docstring is something like

                Called before returning the finished nodes.

                *node* is the reference node if one was created (*is_ref*
                is then true), else the content node.  This method can add
                other nodes and must return a ``(nodes, messages)`` tuple
                (the usual return value of a role function).
        Args:
            document:
                Source document where this ref was defined.
            env:
                Current Sphinx build environment.
            node:
                This role's node.
            is_ref:
                True when this is the reference node, else it's the content
                node.

        Returns:
            A tuple having a list of final nodes for this role and a list
            of system messages if appropriate.
        """
        log.debug(f"Resulting nodes for {self._role_string()}.")
        entry = SphinxGeneralIndexEntry(
            entrytype="single",
            entryname=self.target,
            targetid="",  # targetid=self.rawtext,
            mainname=node.attributes.get("refdoc", ""),
            key=None,
        )
        inode = addnodes.index(entries=[entry])
        node.append(inode)
        return [node], []

    def _role_string(self) -> str:
        if self.has_explicit_title:
            return f":{self.name}:`{self.title} <{self.target}>`"
        else:
            return f":{self.name}:`{self.target}`"


class TerraformDefinitionsIndex(Index):
    name = "definitionsindex"
    localname = "Terraform Definitions Index"
    shortname = localname

    def generate(
        self, docnames: Optional[Iterable[str]] = None
    ) -> Tuple[List[Tuple[str, List[IndexEntry]]], bool]:
        """
        Generate terraform domain index entries.

        Note:
             Entries should be filtered by the docnames provided. To do.

        Args:
            docnames: Restrict source Restructured text documents to these.

        Returns:
            See :meth:`sphinx.domains.Index.generate` for details.
        """
        content_working_copy = defaultdict(list)

        if not isinstance(self.domain, TerraformDomain):
            raise SphinxTerraformError("Incompatible domain class.")

        objects = self.domain.get_objects()

        # generate the expected output, shown below, from the above using the
        # first letter of the npc as a key to group thing
        #
        # name, subtype, docname, anchor, extra, qualifier, description
        #
        # This shows:
        #
        #     D
        #   - **Display Name** *(extra info)* **qualifier:** typ
        #       **Sub Entry** *(extra info)* **qualifier:** typ
        SUBTYPE_NORMAL = 0
        SUBTYPE_WITHSUBS = 1  # noqa: F841
        SUBTYPE_SUB = 2  # noqa: F841

        name_entry: SphinxDomainObjectDescription

        for name_entry in objects:
            first_letter = name_entry.dispname[0].lower()
            content_working_copy[first_letter].append(
                IndexEntry(
                    name_entry.name,
                    SUBTYPE_NORMAL,
                    name_entry.docname,
                    name_entry.anchor,
                    "",  # extra info
                    "",  # qualifier
                    name_entry.type,
                )
            )

        # convert the dict to the sorted list of tuples expected
        content = sorted(content_working_copy.items())

        return content, True


D = TypeVar("D", bound="TerraformDomain")


class SphinxData(NamedTuple):
    identifier: str
    sphinx_obj: SphinxDomainObjectDescription
    module: TerraformModule
    signature: HclSignature
    definition: HclDefinition


class DomainData(TypedDict, total=False):
    version: int
    terraform: Dict[Any, Any]
    sphinx: Dict[str, SphinxData]


class TerraformDomain(Domain):

    name: str = "tf"
    """
    The domain name, short and unique.
    """

    label: str = "Terraform"
    object_types = {
        "resource": ObjType(t_("resource"), "resource"),
        "variable": ObjType(t_("variable"), "variable"),
        "output": ObjType(t_("output"), "output"),
        "module": ObjType(t_("module"), "module"),
        "data": ObjType(t_("data"), "data"),
    }
    directives = {
        "resource": TerraformObjectDirective,
        "variable": TerraformObjectDirective,
        "output": TerraformObjectDirective,
        "module": TerraformObjectDirective,
        "data": TerraformObjectDirective,
    }
    roles = {
        "resource": TerraformCrossReferenceRole(),
        "variable": TerraformCrossReferenceRole(),
        "output": TerraformCrossReferenceRole(),
        "module": TerraformCrossReferenceRole(),
        "data": TerraformCrossReferenceRole(),
    }
    indices: List[Type[Index]] = [TerraformDefinitionsIndex]
    initial_data: DomainData = {  # type: ignore # base class defined the type as "Dict[Any, Any]"
        "terraform": TerraformStore.initial_data(),
        "sphinx": {},
    }

    @classmethod
    def get_instance(
        cls: Type[D], app_or_env: Union[Sphinx, BuildEnvironment]
    ) -> D:
        env = get_env(app_or_env)
        domain = env.get_domain(cls.name)
        return domain  # type: ignore

    @property
    def store(self) -> TerraformStore:
        return TerraformStore(self.data["terraform"])

    @property
    def sphinx_references(self) -> Dict[str, SphinxData]:
        return self.data["sphinx"]  # type: ignore

    def register(
        self,
        module: TerraformModule,
        signature: HclSignature,
    ) -> HclDefinition:
        definition = self.store.register(module, signature, self.env.docname)
        identifier = make_identifier(signature, module)

        sphinx_obj = SphinxDomainObjectDescription(
            name=str(signature),
            dispname=f"{signature.type.value} {signature}",
            type=signature.type.value,
            docname=self.env.docname,
            anchor=identifier,
            priority=0,
        )

        self.sphinx_references[identifier] = SphinxData(
            identifier, sphinx_obj, module, signature, definition
        )

        return definition

    def get_objects(self) -> Iterable[SphinxDomainObjectDescription]:
        """
        Return an iterable of "object descriptions".

        See Also:
             Parent method :meth:`sphinx.domains.Domain.get_objects`.

        Returns:
            Object descriptions are tuples with six items.
            See :class:`sphinx_terraform.sphinxapi.SphinxDomainObjectDescription`.
        """
        for sphinx_data in self.sphinx_references.values():
            yield sphinx_data.sphinx_obj

    def resolve_xref(
        self,
        env: "BuildEnvironment",
        fromdocname: str,
        builder: Builder,
        typ: str,
        target: str,
        node: addnodes.pending_xref,
        contnode: Element,
    ) -> Optional[Element]:
        """
        Resolve the pending_xref *node* with the given *typ* and *target*.

        See also:
            The parent class method docstring is something like

                Resolve the pending_xref *node* with the given *typ* and
                *target*.

                This method should return a new node, to replace the xref
                node, containing the *contnode* which is the markup content
                of the cross-reference.

                If no resolution can be found, None can be returned; the
                xref node will then given to the :event:`missing-reference`
                event, and if that yields no resolution, replaced by *contnode*.

                The method can also raise :exc:`sphinx.environment.NoUri`
                to suppress the :event:`missing-reference` event being emitted.

        Args:
            env:
                Current Sphinx build environment.
            fromdocname:
                Document name where the cross-reference was used.
            builder:
                Current Sphinx builder.
            typ:
                Object type name.
            target:
                Looked up object identifier.
            node:
                Document node for the xref.
            contnode:
                The markup content of the cross-reference.

        If no resolution can be found, ``None`` can be returned;
        the xref node will then given to the ``missing-reference`` event,
        and if that yields no resolution, replaced by contnode.

        Returns:
            A reference node or None if no reference could be resolved.
        """

        def not_found() -> None:
            log.warning(
                "Cross reference to '{!s}' "
                "could not be resolved "
                "within the '{!s}' domain".format(target, self.name)
            )

        candidates = self._resolve_target(target, typ)

        if not candidates:
            not_found()
            return None

        if len(candidates) == 1:
            sphinx_reference = candidates[0]
            return make_refnode(
                builder,
                fromdocname,
                sphinx_reference.sphinx_obj.docname,
                sphinx_reference.sphinx_obj.anchor,
                contnode,
                str(sphinx_reference.signature),
            )

        log.warning(f"Many canditates found for xref to '{typ}' '{target}'.")
        return None

    def resolve_any_xref(
        self,
        env: "BuildEnvironment",
        fromdocname: str,
        builder: "Builder",
        target: str,
        node: addnodes.pending_xref,
        contnode: Element,
    ) -> List[Tuple[str, Element]]:
        candidates = self._resolve_target(target)

        return [
            (
                f"{self.name}:{sphinx_reference.sphinx_obj.type}",
                make_refnode(
                    builder,
                    fromdocname,
                    sphinx_reference.sphinx_obj.docname,
                    sphinx_reference.sphinx_obj.anchor,
                    contnode,
                    str(sphinx_reference.signature),
                ),
            )
            for sphinx_reference in candidates
        ]

    def _resolve_target(
        self, target: str, typename: Optional[str] = None
    ) -> List[SphinxData]:
        match = re.match(
            r"^(?P<module>(\w+\/)*)?((?P<type>[a-z]+)-)?(?P<name>(([\w-]+)(\.([\w-]+))*)?)$",
            target,
        )

        if not match:
            return []

        matcher = self._make_matcher(target, typename)

        return list(filter(matcher, self.sphinx_references.values()))

    def _make_matcher(
        self, target: str, typename: Optional[str]
    ) -> Callable[[SphinxData], bool]:
        match = re.match(
            r"^(?P<module>(\w+\/)*)?((?P<type>[a-z]+)-)?(?P<name>(([\w-]+)(\.([\w-]+))*)?)$",
            target,
        )

        if not match:
            return lambda x: False

        def matcher(sphinx_reference: SphinxData) -> bool:
            matched_module = match.group("module")  # type: ignore # matched checked above
            matched_name = match.group("name")  # type: ignore # matched checked above

            typename_ok = (
                not typename or typename == sphinx_reference.sphinx_obj.type
            )
            module_ok = (
                not matched_module
                or matched_module.strip("/") == sphinx_reference.module.fullname
            )
            name_ok = not matched_name or matched_name in str(
                sphinx_reference.signature
            )

            return all([typename_ok, module_ok, name_ok])

        return matcher



from __future__ import annotations

import os
import shutil

import docutils
import pytest
import sphinx
from sphinx.testing import comparer
from sphinx.testing.path import path

sample_folder = "samples"

collect_ignore = [sample_folder]


@pytest.fixture(scope="session")
def rootdir():
    return path(__file__).parent.abspath() / sample_folder


def pytest_report_header(config):
    header = "libraries: Sphinx-%s, docutils-%s" % (
        sphinx.__display_version__,
        docutils.__version__,
    )
    if hasattr(config, "_tmp_path_factory"):
        header += "\nbase tempdir: %s" % config._tmp_path_factory.getbasetemp()

    return header


def pytest_assertrepr_compare(op, left, right):
    comparer.pytest_assertrepr_compare(op, left, right)


def _initialize_test_directory(session):
    if "SPHINX_TEST_TEMPDIR" in os.environ:
        tempdir = os.path.abspath(os.getenv("SPHINX_TEST_TEMPDIR"))
        print("Temporary files will be placed in %s." % tempdir)

        if os.path.exists(tempdir):
            shutil.rmtree(tempdir)

        os.makedirs(tempdir)


def pytest_sessionstart(session):
    _initialize_test_directory(session)



from __future__ import annotations

import sys
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, NamedTuple, Optional

if TYPE_CHECKING or sys.version_info < (3, 8, 0):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class ExtensionMetadata(TypedDict, total=False):
    """
    Define the optional extension metadata structure.

    See also:
        This is documented in the :ref:`sphinx:ext-metadata` section of
        Sphinx's documentation.
    """

    version: str
    """
    A string that identifies the extension version.

    It is used for extension version requirement checking (see
    :confval:`sphinx:needs_extensions`) and informational purposes.
    If not given, "unknown version" is substituted.
    """

    env_version: int
    """
    An integer that identifies the version of env data structure if the
    extension stores any data to environment.

    It is used to detect the data structure has been changed from last build.
    The extensions have to increment the version when data structure has changed.
    If not given, Sphinx considers the extension does not stores any data
    to environment.
    """

    parallel_read_safe: bool
    """
    A boolean that specifies if parallel reading of source files can be used
    when the extension is loaded.

    It defaults to ``False``, i.e. you have to explicitly specify your extension
    to be parallel-read-safe after checking that it is.
    """

    parallel_write_safe: bool
    """
    A boolean that specifies if parallel writing of output files can be used
    when the extension is loaded.

    Since extensions usually don’t negatively influence the process, this
    defaults to True.
    """


class SphinxEvent(Enum):
    """
    Enumerate Sphinx core events.

    .. note:: Developer, add missing events as needed.

    See also:
        https://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx-core-events
    """

    CONFIG_INITED = "config-inited"
    BUILDER_INITED = "builder-inited"
    ENV_GET_OUTDATED = "env-get-outdated"
    ENV_BEFORE_READ_DOCS = "env-before-read-docs"
    ENV_PURGE_DOC = "env-purge-doc"
    SOURCE_READ = "source-read"
    OBJECT_DESCRIPTION_TRASFORM = "object-description-transform"
    DOCTREE_READ = "doctree-read"
    MISSING_REFERENCE = "missing-reference"
    WARN_MISSING_REFERENCE = "warn-missing-reference"
    DOCTREE_RESOLVED = "doctree-resolved"
    ENV_MERGE_INFO = "env-merge-info"
    ENV_UPDATED = "env-updated"
    ENV_CHECK_CONSISTENCY = "env-check-consistency"
    HTML_COLLECT_PAGES = "html-collect-pages"
    HTML_PAGE_CONTEXT = "html-page-context"
    LINKCHECK_PROCESS_URI = "linkcheck-process-uri"
    BUILD_FINISHED = "build-finished"


class HtmlPage(NamedTuple):
    """
    Expected tuple by Sphinx when collecting HTML pages.
    """

    pagename: str
    context: Dict[str, Any]
    templatename: str


class SphinxDomainObjectDescription(NamedTuple):
    """
    Identify something of interest for search results, for instance.

    Sphinx uses a lot of tuples.  This one is documented at
    :meth:`sphinx.domains.Domain.get_objects`.
    """

    name: str
    """
    Fully qualified name.
    """

    dispname: str
    """
    Name to display when searching/linking.
    """

    type: str
    """
    Object type name, which may or may not match role and directive names.
    """

    docname: str
    """
    Document name where it is to be found.
    """

    anchor: str
    """
    The anchor name for the object.
    """

    priority: int
    """
    How "important" the object is (determines placement in search results).

    One of:

    ``1``
        Default priority (placed before full-text matches).
    ``0``
        Object is important (placed before default-priority objects).
    ``2``
        Object is unimportant (placed after full-text matches).
    ``-1``
        Object should not show up in search at all.
    """


class SphinxGeneralIndexEntry(NamedTuple):
    """
    An entry in the general index.

    This tuple is documented in :meth:`sphinx.domains.Index.generate`.
    """

    entrytype: str
    """
    One of "single", "pair", "double", "triple".
    """

    entryname: str
    """
    Index entry name.

    As an example, in the case of glossary definitions with multiple terms
    (synonyms, for instance), each term gets its index entry and the entryname
    is the term.
    """

    targetid: str

    mainname: str

    key: Optional[str]
    """
    Categorization characters (usually a single character) for general index page.
    """



from __future__ import annotations

import re

import pytest
from sphinx.application import Sphinx


@pytest.mark.sphinx(testroot="basic")
def test_tf(built_index_html: str, status, warning):
    html = (
        r"<h1>test-basic<a "
        r'class="headerlink" '
        r'href="#test-basic" '
        r'title="Permalink to this headline">¶'
        r"</a></h1>"
    )

    assert re.search(html, built_index_html, re.DOTALL)


@pytest.mark.sphinx(testroot="md-comment-from-md-document")
def test_md_in_md(built_index_html: str, status, warning):
    html = (
        r"<p>Terraform documentation in Markdown</p>\s*"
        r"<p>.+<a .+>hyperlink</a></p>"
    )

    assert re.search(html, built_index_html, re.DOTALL)


@pytest.mark.sphinx(testroot="rst-comment-from-rst-document")
def test_rst_in_rst(built_index_html: str, status, warning):
    html = (
        r"<p>Terraform documentation in RestructuredText</p>\s*"
        r"<p>.+<a .+>hyperlink</a></p>"
    )

    assert re.search(html, built_index_html, re.DOTALL)


@pytest.mark.sphinx(testroot="md-directive-in-rst-document")
def test_md_in_rst(built_index_html: str, status, warning):
    html = (
        r"<p>Terraform documentation in Markdown</p>\s*"
        r"<p>.+<a .+>hyperlink</a></p>"
    )

    assert re.search(html, built_index_html, re.DOTALL)


@pytest.mark.sphinx(
    testroot="md-directive-in-rst-document",
    confoverrides={"terraform_comment_markup": "rst"},
)
def test_md_in_rst_force_conf(built_index_html: str, status, warning):
    html = (
        r"<p>Terraform documentation in Markdown</p>\s*"
        r"<p>.+<a .+>hyperlink</a></p>"
    )

    assert re.search(html, built_index_html, re.DOTALL)


@pytest.mark.sphinx(testroot="rst-directive-in-md-document")
def test_rst_in_md(built_index_html: str, status, warning):
    html = (
        r"<p>Terraform documentation in RestructuredText</p>\s*"
        r"<p>.+<a .+>hyperlink</a></p>"
    )

    assert re.search(html, built_index_html, re.DOTALL)


@pytest.mark.sphinx(
    testroot="rst-directive-in-md-document",
    confoverrides={"terraform_comment_markup": "md"},
)
def test_rst_in_md_force_conf(built_index_html: str, status, warning):
    html = (
        r"<p>Terraform documentation in RestructuredText</p>\s*"
        r"<p>.+<a .+>hyperlink</a></p>"
    )

    assert re.search(html, built_index_html, re.DOTALL)


@pytest.mark.sphinx(testroot="config-terraform_comment_markup-md")
def test_config_terraform_comment_markup_md(
    built_index_html: str, status, warning
):
    html = (
        r"<p>Terraform documentation in Markdown</p>\s*"
        r"<p>.+<a .+>hyperlink</a></p>"
    )

    assert re.search(html, built_index_html, re.DOTALL)


@pytest.mark.sphinx(testroot="config-terraform_comment_markup-rst")
def test_config_terraform_comment_markup_rst(
    built_index_html: str, status, warning
):
    html = (
        r"<p>Terraform documentation in RestructuredText</p>\s*"
        r"<p>.+<a .+>hyperlink</a></p>"
    )

    assert re.search(html, built_index_html, re.DOTALL)


@pytest.fixture()
def built_index_html(app: Sphinx) -> str:
    app.builder.build_all()

    content = (app.outdir / "index.html").read_text()
    return content



from __future__ import annotations

from pathlib import Path

from setuptools import find_packages, setup

extras_require = {
    "test": ["nox"],
    "markdown": ["myst_parser", "docutils>=0.16"],
    "docs": [
        "myst_parser",
        "docutils>=0.16",
        "sphinx_autodoc_typehints",
        "sphinx_paramlinks",
        "sphinx_copybutton",
        "sphinxcontrib_bibtex",
        "sphinxcontrib_programoutput",
        "sphinx_tabs",
        "pydata-sphinx-theme",
    ],
}
extras_require["all"] = set(
    dependency
    for extra_dependencies in extras_require.values()
    for dependency in extra_dependencies
)

this_directory = Path(__file__).parent
requirements_in = this_directory.joinpath("requirements.in")


if __name__ == "__main__":
    setup(
        name="sphinx-terraform",
        version="0.1",
        description="A Sphinx extension for documenting Terraform modules.",
        long_description=(this_directory / "README.rst").read_text(),
        long_description_content_type="text/x-rst",
        url="https://cblegare.gitlab.io/sphinx-terraform",
        author="Charles Bouchard-Légaré",
        author_email="charlesbouchardlegare@gmail.com",
        license="BSD-2-Clause-Patent",
        project_urls={
            "Documentation": "https://cblegare.gitlab.io/sphinx-terraform",
            "Source": "https://gitlab.com/cblegare/sphinx-terraform",
            "Issue Tracker": "https://gitlab.com/cblegare/sphinx-terraform/-/issues",
        },
        classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            "Development Status :: 3 - Alpha",
            # Indicate who your project is intended for
            "Intended Audience :: Developers",
            "Intended Audience :: Information Technology",
            "Intended Audience :: System Administrators",
            "Topic :: Documentation",
            "Topic :: Documentation :: Sphinx",
            "Framework :: Sphinx",
            "Framework :: Sphinx :: Extension",
            # Pick your license as you wish (should match "license" above)
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Operating System :: OS Independent",
            "Typing :: Typed",
        ],
        keywords="hcl tf terraform sphinx iac cloud infrastructure",
        packages=find_packages(where="src"),
        package_dir={"": "src"},
        python_requires=">=3.7",
        install_requires=requirements_in.read_text().splitlines(),
        extras_require=extras_require,
        include_package_data=True,
    )



from __future__ import annotations

from pathlib import Path
import shutil
from typing import List

import nox

supported_python_versions = ["3.7", "3.8", "3.9", "3.10"]

default_python = sorted(supported_python_versions)[-1]

repo_root = Path(__file__).parent


@nox.session(reuse_venv=True)
def build(session):
    session.install("build", "twine")
    session.run("python", "-m", "build")
    session.run("python", "-m", "twine", "check", "--strict", "dist/*")


@nox.session(reuse_venv=True)
def requirements(session):
    shutil.copyfile(
        repo_root / "requirements.in", repo_root / "requirements.txt"
    )


@nox.session(reuse_venv=True)
def format(session):
    session.install("black", "isort")
    session.run("black", ".")
    session.run("isort", ".")


@nox.session(python=supported_python_versions)
def fulltest(session):
    session.notify(f"pytest-{session.python}")
    session.notify(f"doctest-{session.python}")
    session.notify(f"e2etest-{session.python}")
    session.notify(f"typing-{session.python}")


@nox.session(python=supported_python_versions)
def pytest(session):
    session.install(".[markdown]", "pytest", "coverage[toml]")
    session.run(*_coverage_cmd(session.name, ["pytest", "test"]))


@nox.session(python=supported_python_versions)
def doctest(session):
    session.install(".", "pytest", "coverage[toml]")
    session.run(
        *_coverage_cmd(session.name, ["pytest", "--doctest-modules", "src"])
    )


@nox.session(python=supported_python_versions)
def e2etest(session):
    session.install(".[docs]", "coverage[toml]")
    session.run(*_python_cmd(_sphinx_apidoc_modulecmd()))
    session.run(
        *_coverage_cmd(session.name, _sphinx_build_modulecmd("build/sphinx"))
    )


@nox.session(reuse_venv=True)
def lint(session):
    session.install("flake8", "black", "isort")
    session.run("black", "--check", "--diff", ".")
    session.run("isort", "--check", "--diff", ".")
    session.run("flake8")


@nox.session(reuse_venv=True)
def docstringstyle(session):
    session.install("pydocstyle", "toml")
    session.run("pydocstyle", "src")


@nox.session(reuse_venv=True, python=supported_python_versions)
def typing(session):
    session.install(
        ".[markdown]",
        "mypy",
        "lxml",
        "docutils-stubs",
        "importlib_metadata",
    )
    session.run("mypy", "--python-version", session.python, "src")


@nox.session(reuse_venv=True)
def coverage(session):
    session.install(".", "coverage[toml]")
    session.run("coverage", "combine")
    session.run("coverage", "xml")
    session.run("coverage", "html")
    session.run("coverage", "report")


@nox.session(python=default_python, reuse_venv=True)
def docs(session):
    session.install(".[docs]")
    session.run(*_python_cmd(_sphinx_apidoc_modulecmd()))
    session.run(*_python_cmd(_sphinx_build_modulecmd("build/sphinx")))


def _sphinx_apidoc_modulecmd() -> List[str]:
    return [
        "sphinx.ext.apidoc",
        "-o",
        "docs/apidoc",
        "--force",
        "--no-toc",
        "--separate",
        "--module-first",
        "src",
    ]


def _sphinx_build_modulecmd(build_root: str) -> List[str]:
    return [
        "sphinx",
        "-b",
        "html",
        "-d",
        f"{build_root}/doctrees",
        "-E",
        "-n",
        "-W",
        "--keep-going",
        "-T",
        "docs",
        f"{build_root}/html",
    ]


def _python_cmd(modulecmd: List[str]) -> List[str]:
    return ["python", "-m", *modulecmd]


def _coverage_cmd(context: str, modulecmd: List[str]) -> List[str]:
    return [
        "python",
        "-m",
        "coverage",
        "run",
        f"--context={context}",
        "-m",
        *modulecmd,
    ]



from __future__ import annotations

import os
import sys
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Type, Union

from sphinx.util.logging import getLogger

from sphinx_terraform import get_builder, get_highlighter
from sphinx_terraform.i18n import t_
from sphinx_terraform.markup import TerraformDomain
from sphinx_terraform.sphinxapi import HtmlPage
from sphinx_terraform.terraform import TerraformModule, make_identifier

if TYPE_CHECKING or sys.version_info < (3, 8, 0):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


log = getLogger(__name__)


class HtmlWriter:
    def __init__(self, domain: TerraformDomain, output_dirname: Path):
        self._domain = domain
        self._output_dirname = Path(output_dirname)

        self._highlighter = get_highlighter(domain.env)

    def make_pagename(self, tf_file: Path) -> Path:
        page_name = self._output_dirname.joinpath(
            self.get_symbolic_path(tf_file)
        )
        return page_name

    def gen_sourcecode_html(self, tf_file: Path) -> HtmlPage:
        code = self._domain.store.get_code(tf_file)
        pagename = self.make_pagename(tf_file)
        highlighted_code = self.highlight_hcl_code(code)

        annotated_highlighted_code = self._add_backlinks_to_documentation(
            highlighted_code, tf_file
        )

        log.debug(f"Emitting HTML for {tf_file}")
        context = {
            "title": self.get_symbolic_path(tf_file),
            "body": (
                t_("<h1>Source code for %s</h1>")
                % self.get_symbolic_path(tf_file)
                + "\n"
                + "\n".join(annotated_highlighted_code)
            ),
        }
        return HtmlPage(str(pagename), context, "page.html")

    def gen_root_module_index(self) -> HtmlPage:
        tree_of_modules = self._get_tree_of_modules_and_files()

        pagename = self._output_dirname.joinpath("index")

        def file_to_url(tf_file: Path) -> str:
            return self._urito(
                pagename,
                self._output_dirname.joinpath(self.get_symbolic_path(tf_file)),
            )

        html_list = list(
            self._tree_of_modules_to_html_list(tree_of_modules, file_to_url)
        )

        context = {
            "title": t_("Overview: module code"),
            "body": (
                t_("<h1>All modules for which code is available</h1>")
                + "\n"
                + "\n".join(html_list)
            ),
        }
        return HtmlPage(str(pagename), context, "page.html")

    def highlight_hcl_code(self, code: List[str]) -> List[str]:
        lexer = "tf"

        highlighted = self._highlighter.highlight_block(
            os.linesep.join(code), lexer, linenos=False
        )
        highlighted_lines = highlighted.splitlines()
        before, after = highlighted_lines[0].split("<pre>")
        highlighted_lines[0:1] = [before + "<pre>", after]

        return highlighted_lines

    def get_symbolic_path(self, tf_file: Path) -> Path:
        """
        Return the path to a file name starting with the root module's name.
        """
        module = self._domain.store.get_module(tf_file)

        return Path(module.fullname, tf_file.relative_to(module.path))

    def _add_backlinks_to_documentation(
        self,
        highlighted_code: List[str],
        tf_file: Path,
    ) -> List[str]:
        pagename = self.make_pagename(tf_file)
        for line_number, line_of_code in enumerate(highlighted_code):
            if line_of_code.startswith("<span"):
                first_line_of_code = line_number
                break
        else:
            first_line_of_code = 0

        module = self._domain.store.get_module(tf_file)

        for signature, entry in self._domain.store.get_definitions(
            tf_file=tf_file
        ).items():
            for docname_where_documented in entry.usages:
                backlink = (
                    f"{self._urito(pagename, docname_where_documented)}"
                    "#"
                    f"{make_identifier(signature, module)}"
                )
                definition_starting_line = (
                    entry.code.start_position.line + first_line_of_code
                )
                definition_ending_line = (
                    entry.code.end_position.line + first_line_of_code
                )
                highlighted_code[definition_starting_line] = "".join(
                    [
                        f'<div class="viewcode-block" id="{make_identifier(signature)}">',
                        f'<a class="viewcode-back" href="{backlink}">',
                        f"{t_('[docs]')}</a>{highlighted_code[definition_starting_line]}",
                    ]
                )
                highlighted_code[definition_ending_line] += "</div>"
        return highlighted_code

    def _urito(
        self,
        from_page: Union[str, Path],
        to_page: Union[str, Path],
        typ: Optional[str] = None,
    ) -> str:
        return get_builder(self._domain.env).get_relative_uri(
            str(from_page), str(to_page), typ  # type: ignore # typ is wrongly typed upstream
        )

    def _get_tree_of_modules_and_files(self) -> TreeOfModulesAndFiles:
        known = set()
        result: TreeOfModulesAndFiles = {}

        documented_files = sorted(self._domain.store.get_documented_files())

        for tf_file in documented_files:
            module = self._domain.store.get_module(tf_file)
            stack = result
            if module in known:
                continue
            known.add(module)

            for part in module.fullname.split("/")[1:]:
                if part not in stack:
                    stack[part] = {}  # type: ignore
                stack = stack[part]  # type: ignore

            stack["__tf_files__"] = list(
                self._domain.store.get_documented_files(module)
            )
            stack["__tf_module__"] = module

        return result

    def _tree_of_modules_to_html_list(
        self,
        _tree: TreeOfModulesAndFiles,
        file_to_url: Callable[[Path], str],
    ) -> List[str]:
        html_parts: List[str] = []
        html = Html(html_parts)

        def _recurse(_tree: TreeOfModulesAndFiles) -> None:
            with html.tag("ul"):
                module = _tree["__tf_module__"]
                with html.tag("li"):
                    with html.tag("strong"):
                        html.append(str(module))

                files = _tree.get("__tf_files__", [])
                if files:
                    with html.tag("ul"):
                        for filepath in files:
                            with html.tag("li"):
                                with html.tag("a", href=file_to_url(filepath)):
                                    html.append(
                                        str(self.get_symbolic_path(filepath))
                                    )

                for subtree in [
                    child
                    for key, child in _tree.items()
                    if not isinstance(key, str) and child
                ]:
                    _recurse(subtree)  # type: ignore

        _recurse(_tree)
        return html_parts


class TreeOfModulesAndFiles(TypedDict, total=False):
    __tf_files__: List[Path]
    __tf_module__: TerraformModule


class Html:
    class _inner(object):
        def __init__(self, tagger: Html, tag: str, **attributes: str):
            self.tagger = tagger
            self.tag = tag
            self.attributes = attributes

        def __enter__(self) -> None:
            if self.attributes:
                attribute_str = " ".join(
                    f'{key}="{value}"' for key, value in self.attributes.items()
                )
                self.tagger.append(f"<{self.tag} {attribute_str}>")
            else:
                self.tagger.append(f"<{self.tag}>")
            self.tagger.indent_count += 1

        def __exit__(
            self,
            exctype: Optional[Type[BaseException]],
            excvalue: Optional[BaseException],
            traceback: Optional[TracebackType],
        ) -> None:
            self.tagger.indent_count -= 1
            self.tagger.append(f"</{self.tag}>")

    def __init__(
        self,
        container: List[str],
        initial_indent_count: int = 0,
        indent_increment: str = "  ",
    ):
        self.container = container
        self.indent_count = initial_indent_count
        self.indent_increment = indent_increment

    def tag(self, tag: str, **attributes: str) -> Any:
        return self._inner(self, tag, **attributes)

    def append(self, some_html: str) -> None:
        self.container.append(
            f"{self.indent_increment * self.indent_count}{some_html}"
        )
